#import "BuiltInDeviceView.h"
#import "RInterpreter.h"
#import <InstantMessage/IMService.h>
#import <InstantMessage/IMAVManager.h>
#import <CoreVideo/CoreVideo.h>

@implementation BuiltInDeviceView

- (id)initWithFrame:(NSRect)frame {
    self = [super initWithFrame:frame];
    if (self) {
		dev = nil;
		[[NSNotificationCenter defaultCenter] addObserver:self
										   selector:@selector(evalDidFinish:) name:RInterpreterDidFinishEvaluating object:nil];
		locateLayer = NULL;
		doodleLayer = NULL;
		context     = NULL;
		dirty       = NO;
		sendFrame   = NO;
		doodle      = NO;		
		cachedImage = nil;
		cachedSize  = NSMakeSize(0,0);
    }
    return self;
}


- (void)finalize {
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	if(NULL != locateLayer)
		CGLayerRelease(locateLayer);
	[super finalize];
}

- (void)drawInBounds:(NSRect)b {
	//Only draw in a single thread.
	@synchronized(self) {
		//If our we're dirty we need to update our cached image.
		if(YES == dirty || nil == cachedImage) {
			if(cachedSize.width != b.size.width || cachedSize.height != b.size.height) {
				cachedImage = [self bitmapImageRepForCachingDisplayInRect:b];
				cachedSize  = b.size;
			}
			[NSGraphicsContext saveGraphicsState];
			[NSGraphicsContext setCurrentContext:[NSGraphicsContext graphicsContextWithBitmapImageRep:cachedImage]];
			[[NSColor whiteColor] set];
			NSRectFill(b);
			if(nil != dev) 
				[dev drawInRect:b];
			[NSGraphicsContext restoreGraphicsState];
			dirty = NO;
		}	
		[cachedImage drawInRect:b];
		//Draw the locator layer if needed
		CGContextRef ctx = [[NSGraphicsContext currentContext] graphicsPort];
		CGRect       br  = CGRectMake(b.origin.x,b.origin.y,b.size.width,b.size.height);
		if(NULL != locateLayer)
			CGContextDrawLayerInRect(ctx,br,locateLayer);
		if(NULL != doodleLayer)
			CGContextDrawLayerInRect(ctx,br,doodleLayer);
	}
}

- (void)drawRect:(NSRect)rect {
	if(NO == [self inLiveResize]) 
		[self drawInBounds:[self bounds]];
}

- (void)viewDidEndLiveResize {
	[dev setSize:[self bounds].size];
	if(NULL != locateLayer) {
		CFRelease(locateLayer);
		locateLayer = NULL;
	}
	if(NULL != doodleLayer) {
		CFRelease(doodleLayer);
		doodleLayer = NULL;
	}
		
	[dev redraw];
	dirty = YES;
	sendFrame = YES;
	[self setNeedsDisplay:YES];
}


//This is too granular for most graphics flushing purposes. We pay a price ("interactivity") for a big
//speed gain by ignoring it.
- (void)graphicsDeviceDidDraw:(id)aDevice {
	dirty = YES;
}

- (void)evalDidFinish:(id)sender {
	if(YES == dirty) {
		[self display];
		sendFrame = YES;
	}
}

- (void)graphicsDeviceNewPage:(id)aDevice {
	if(NULL != locateLayer) {
		CFRelease(locateLayer);
		locateLayer = NULL;
	}
	if(NULL != doodleLayer) {
		CFRelease(doodleLayer);
		doodleLayer = NULL;
	}
	dirty = YES;
	sendFrame = YES;
	[self setNeedsDisplay:YES];
}

- (void)graphicsDeviceClose:(id)aDevice {
	[[self window] orderOut:self];
}

- (void)setDeviceDescription:(void*)devDesc {
	NSSize s = [self bounds].size;
	dev = [RLayerGraphicsDevice deviceWithSize:NSMakeSize(s.width,s.height) deviceDesc:devDesc delegate:self];
	dev.scaleFactor = [[self window] userSpaceScaleFactor];
}

- (BOOL)graphicsDeviceCanResize:(id)aDevice { return YES; }

- (NSSize)graphicsDeviceSize:(id)aDevice {
//	NSLog(NSStringFromRect([self bounds]));
	NSSize s = [self bounds].size;
	return NSMakeSize(s.width,s.height);
}

- (BOOL)graphicsDeviceCanLocate:(id)aDevice {
	return TRUE;
}
- (void)graphicsDeviceWillLocate:(id)aDevice {
	NSSize s = [self bounds].size;
	locateLayer = CGLayerCreateWithContext(CGDisplayGetDrawingContext([[[[NSScreen mainScreen] deviceDescription] objectForKey:@"NSScreenNumber"] unsignedIntegerValue]),
									 CGSizeMake(s.width,s.height),NULL);
	context = CGLayerGetContext(locateLayer);
	locate = YES;
}

- (void)graphicsDeviceDidLocate:(id)aDevice {
	if(NULL != locateLayer) {
		CGLayerRelease(locateLayer);
		locateLayer = NULL;
	}
	locate = NO;
}
- (void)graphicsDevice:(id)aDevice showCoords:(NSPoint)rPoint atPoint:(NSPoint)aPoint {
	CGContextBeginPath(context);
	CGContextSetRGBFillColor(context,0.0,0.0,0.0,0.1);
	CGContextSetRGBStrokeColor(context,0.0,0.0,0.0,1.0);	
	CGContextAddEllipseInRect(context,CGRectMake(aPoint.x-5,aPoint.y-5,10,10));
	CGContextDrawPath(context,kCGPathFillStroke);
	sendFrame = YES;
	[self setNeedsDisplay:YES];
}

- (BOOL)acceptsFirstResponder { return YES; }


- (void)drawDoodlePoint:(NSPoint)aPoint {
	if(NULL == doodleLayer) {
		NSSize s = [self bounds].size;
		doodleLayer = CGLayerCreateWithContext(
				CGDisplayGetDrawingContext([[[[NSScreen mainScreen] deviceDescription] objectForKey:@"NSScreenNumber"] unsignedIntegerValue]),
				CGSizeMake(s.width,s.height),NULL);
	}
	

/*	CGContextRef ctx = CGLayerGetContext(doodleLayer);
	CGContextBeginPath(ctx);
	CGContextSetRGBFillColor(ctx,1.0,0.0,0.0,1.0);
	CGContextSetRGBStrokeColor(ctx,1.0,0.0,0.0,1.0);	
	CGContextAddEllipseInRect(ctx,CGRectMake(aPoint.x-2.5,aPoint.y-2.5,5,5));
	CGContextDrawPath(ctx,kCGPathFillStroke);*/
	lastDoodlePoint = aPoint;

	sendFrame = YES;
	[self setNeedsDisplay: YES];
}

- (void)drawDoodleLine:(NSPoint)aPoint {
	if(NULL == doodleLayer) {
		NSSize s = [self bounds].size;
		doodleLayer = CGLayerCreateWithContext(
				CGDisplayGetDrawingContext([[[[NSScreen mainScreen] deviceDescription] objectForKey:@"NSScreenNumber"] unsignedIntegerValue]),
				CGSizeMake(s.width,s.height),NULL);
	}
	

	CGContextRef ctx = CGLayerGetContext(doodleLayer);
	CGContextBeginPath(ctx);
	CGContextSetRGBStrokeColor(ctx,1.0,0.0,0.0,1.0);	
	CGContextSetLineCap(ctx,kCGLineCapRound);
	CGContextSetLineWidth(ctx, 5);
	CGContextMoveToPoint(ctx, lastDoodlePoint.x, lastDoodlePoint.y);
	CGContextAddLineToPoint(ctx, aPoint.x, aPoint.y);
	CGContextDrawPath(ctx,kCGPathFillStroke);
	lastDoodlePoint = aPoint;
	sendFrame = YES;
	[self setNeedsDisplay: YES];

}

- (void)mouseDown:(NSEvent*)event {
//	NSLog(@"mouseDown:%@",[event description]);
	if(YES == locate) {
		NSPoint eventLocation = [event locationInWindow];
		if([event clickCount] > 1)
			[dev stopLocating];
		else
			[dev setLocatePoint:[self convertPoint:eventLocation fromView:nil]];
	} else {
		[self drawDoodlePoint:[self convertPoint:[event locationInWindow] fromView:nil]];
		doodle = YES;
	}
	[super mouseDown:event];
}

- (void)mouseDragged:(NSEvent*)event {
	if(YES == doodle)
		[self drawDoodleLine:[self convertPoint:[event locationInWindow] fromView:nil]];
	[super mouseMoved:event];
}

- (void)mouseUp:(NSEvent*)event {
	doodle = NO;
	[super mouseUp:event];
}

#pragma mark -
#pragma mark IMAVManager delegate methods

- (void)activateIMLink:(id)sender {
	sendFrame = YES;
}
- (void)deactivateIMLink:(id)sender {
	sendFrame = NO;
}

- (void)getPixelBufferPixelFormat:(OSType*)pixelFormatOut {
	*pixelFormatOut = k32ARGBPixelFormat;
}

- (BOOL)renderIntoPixelBuffer:(CVPixelBufferRef)buffer forTime:(CVTimeStamp*)timeStamp {
	if(NO == sendFrame) return NO;
	@synchronized(self) {
		unsigned char *baseAddress;
		
		CVPixelBufferLockBaseAddress(buffer,0);
		baseAddress = (unsigned char *)CVPixelBufferGetBaseAddress(buffer);
		CGContextRef bitmap = CGBitmapContextCreate(baseAddress,CVPixelBufferGetWidth(buffer),CVPixelBufferGetHeight(buffer),8,CVPixelBufferGetBytesPerRow(buffer),
													CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB),kCGImageAlphaPremultipliedFirst);
		[NSGraphicsContext saveGraphicsState];
		[NSGraphicsContext setCurrentContext:[NSGraphicsContext graphicsContextWithGraphicsPort:bitmap flipped:NO]];

		NSRect newFrame  = NSMakeRect(0,0,CVPixelBufferGetWidth(buffer),CVPixelBufferGetHeight(buffer));
/*		
		[[NSColor blackColor] set];		
		NSRectFill(newFrame);
		NSRect myFrame   = [self frame];
		float  aspect   = myFrame.size.width/myFrame.size.height;
		if(myFrame.size.width > myFrame.size.height) {
			float newWidth = (newFrame.size.height*aspect);
			newFrame.size.width -= newWidth;
			newFrame.origin.x += newWidth/2.0;
		} else {
		}*/
		[self drawInBounds:newFrame];
		[NSGraphicsContext restoreGraphicsState];
		CGContextRelease(bitmap);
		CVPixelBufferUnlockBaseAddress(buffer,0);
		sendFrame = NO;
		return YES;
	}
	//Here to make gcc happy, the @synchronized keyword should always take over.
	return NO;
}


@end
